#!/usr/bin/env python3
""" Data Structures and Algorithms for CL 3, Assignment 3
    See <https://dsacl3-2023.github.io/a3/> for detailed
    instructions.

    <Please insert your name and the honor code here.>
"""

import json
from numpy.linalg import norm
from numpy import array


class Embedding():
    def __init__(self, word, embedding):
        """
        Helper class representing a word embedding.

        Args:
            word (str): a word
            embedding (vector): its embedding
        """
        self.word = word
        self.embedding = array(embedding)
        
    def distance(self, other):
        """
        Return Euclidean distance from another embedding.
        """
        return norm(self.embedding - other.embedding)
    
    def __eq__(self, other):
        """
        Define equality for this class.
        """
        return self.word == other.word
    
    def __str__(self):
        """
        Return string representation.
        """
        return self.word
    
    def __repr__(self):
        """
        Return string representation when printed inside a list.
        """
        return self.__str__()


class BinaryHeap():

    def __init__(self, key=lambda x: x):
        """
        Binary max-heap impementation using an array (list) as storage.

        Args:
            key (function, optional): sorting criterion, just like in the
            python method `sorted()`. Defaults to `lambda x: x` (the object
            itself).
        """
        self.H = []
        self.key = key
        
    def get_parent(self, node):
        """
        Return index of parent node. Return `None` if the
        given node is the root.
        
        Args:
            node (int): index of child node.
        """
        # TODO Exercise 1.1
        pass
    
    def get_max_child(self, node):
        """
        Return index of child node with the bigger label.
        Return `None` if the node does not have any children.
        
        Args:
            node (int): index of parent node.
        """
        # TODO Exercise 1.2
        pass
        
    def push(self, label):
        """
        Add a new node with label to the heap according to
        the `key` provided in the constructor.
        """
        # TODO Exercise 1.3
        pass
            
    def pop(self):
        """
        Remove and return the maximum value from the heap.
        """
        # TODO Exercise 1.4
        pass
    
    def _to_str(self, node=0):
        """
        Recursive helper for the `__str__()` method.
        """
        left = (node*2) + 1
        right = (node*2) + 2
        if right >= len(self.H):
            if left >= len(self.H):
                return str(self.H[node])
            else:
                return '(' + str(self.H[node]) + ' ' + str(self.H[left]) + ')'
        else:
            return '(' + str(self.H[node]) + ' ' \
                    + self._to_str(left) + ' ' + self._to_str(right) + ')'
            
    def __str__(self):
        """
        Convert heap to a string in the Penn Treebank format.
        """
        return self._to_str()
    
    def __iter__(self):
        """
        Initialize iteration when heap is called as an iterable.
        """
        self.iter = 0
        return self
    
    def __next__(self):
        """
        Return next value when heap is called as an iterable.
        """
        if self.iter == len(self.H):
            raise StopIteration
        i = self.iter
        self.iter += 1
        return self.H[i]


def find_similar(target_embedding, embeddings, k=3):
    """
    Find the `k` words that are the most similar to a given target word.

    Args:
        target_embedding (Embedding): target embedding.
        embeddings (list[Embedding]): candidate embeddings.
        k (int, optional): number of output embeddings. Defaults to 3.

    Returns:
        list[str]: top `k` most similar words.
    """
    # TODO Exercise 2
    pass


if __name__ == "__main__":
    
    # Load in {word: embedding} dictionary
    word2embedding = json.load(open("food.json", "r", encoding="utf-8"))
    
    # Pick a target word
    target_word = "pasta"
    if target_word in word2embedding:
        target_embedding = Embedding(target_word, word2embedding[target_word])
    else:
        print("Word embedding not available. Try another word.")
        exit(0)
    
    # Construct Embedding objects from dictionary
    embeddings = list()
    for word in word2embedding:
        if word != target_word:  # skip target
            embeddings.append(Embedding(word, word2embedding[word]))

    # Print most similar words to target
    top_words = find_similar(target_embedding, embeddings)
    top_words_print = [str(i+1)+". "+str(w) for i, w in enumerate(top_words)]
    print(f"\nMost similar to \"{target_word}\":\n\t" + "\n\t".join(top_words_print) + "\n")
